/**
 * Copyright (c) 2015 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 *
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#if 1
#include <stdbool.h>
#include <stdint.h>
#include <string.h>

#include "eeprom_simulator.h"
#include "nrf_drv_twis.h"
#include "nrf_gpio.h"
#include "app_util_platform.h"
#include "nrf_dfu_serial.h"
#include <string.h>
#include "app_util_platform.h"
#include "nrf_dfu_transport.h"
#include "nrf_dfu_req_handler.h"
#include "nrf_log.h"
#include "nrf_dfu_handling_error.h"
#include "nrf_balloc.h"


#define NRF_SERIAL_OPCODE_SIZE          (sizeof(uint8_t))
#define NRF_UART_MAX_RESPONSE_SIZE_SLIP (2 * NRF_SERIAL_MAX_RESPONSE_SIZE + 1)
#define RX_BUF_SIZE                     (64) //to get 64bytes payload
#define OPCODE_OFFSET                   (sizeof(uint32_t) - NRF_SERIAL_OPCODE_SIZE)
#define DATA_OFFSET                     (OPCODE_OFFSET + NRF_SERIAL_OPCODE_SIZE)
#define UART_SLIP_MTU                   (2 * (RX_BUF_SIZE + 1) + 1)
#define BALLOC_BUF_SIZE                 ((CEIL_DIV((RX_BUF_SIZE+OPCODE_SIZE),sizeof(uint32_t))*sizeof(uint32_t)))

NRF_BALLOC_DEF(m_payload_pool, (UART_SLIP_MTU + 1), NRF_DFU_SERIAL_UART_RX_BUFFERS);

static uint8_t m_rx_byte;

static nrf_dfu_serial_t m_serial = {0};
static slip_t m_slip = {0};
static uint8_t m_rsp_buf[NRF_UART_MAX_RESPONSE_SIZE_SLIP];


#define MAX_RESPONSE_SIZE               (1+1+3*4)
static i2c_dfu_t m_dfu;
static bool m_active;

DFU_TRANSPORT_REGISTER(nrf_dfu_transport_t const dfu_trans) =
{
    .init_func  = i2c_dfu_transport_init,
    .close_func = i2c_dfu_transport_close
};

NRF_BALLOC_DEF(m_payload_pooll, (UART_SLIP_MTU + 1), NRF_DFU_SERIAL_UART_RX_BUFFERS);

/**
 * @ingroup eeprom_simulator
 * @defgroup eeprom_simulator_ivars EEPROM simulator internal variables
 *
 * Internal variables required for the module to work.
 * @{
 */

/**
 * @brief Memory array.
 *
 * Place that will simulate the memory.
 */
static uint8_t m_memory[EEPROM_SIM_SIZE];

/**
 * @brief Current memory address.
 *
 * Memory pointer for any operation that will be processed on memory array.
 */
static uint16_t m_addr = 0;

/**
 * @brief Receive buffer.
 *
 * Receiving buffer must contain the address and EEPROM_SIM_SEQ_WRITE_MAX_BYTES bytes of data.
 */
 
//static uint8_t m_rxbuff[EEPROM_SIM_ADDRESS_LEN_BYTES + EEPROM_SIM_SEQ_WRITE_MAX_BYTES];
static uint8_t m_rxbuff[128+1];
static const uint8_t m_length = sizeof(m_rxbuff);        /**< Transfer length. */
static uint8_t       m_tx_buf[MAX_RESPONSE_SIZE];          /**< TX buffer. */
/**
 * @brief Internal error flag.
 *
 * This flag is set if any communication error is detected.
 * It can be cleared by calling the @ref eeprom_simulator_error_get function.
 */
static bool m_error_flag;

/**
 * @brief TWIS instance.
 *
 * TWIS driver instance used by this EEPROM simulator.
 */
static const nrf_drv_twis_t m_twis = NRF_DRV_TWIS_INSTANCE(EEPROM_SIM_TWIS_INST);

/**
 * @brief EEPROM simulator configuration.
 *
 * When the emulator is initialized, the configuration is stored here.
 */
static eeprom_sim_config_t m_config;


/** @} */

/**
 * @ingroup eeprom_simulator
 * @defgroup eeprom_simulator_ifunc EEPROM simulator internal functions
 *
 * Internal auxiliary functions.
 * @{
 */
/**
 * @brief Set the current address pointer.
 *
 * Sets address for the next operation on the memory array.
 * @param addr Address to set.
 */
static void ees_setAddr(uint16_t addr)
{
		m_addr = addr;
}

/**
 * @brief Perform a write operation on the memory array.
 *
 * Write a single byte into the memory array using @ref m_addr as a write pointer.
 * @param data Data to be written.
 */

int num= 0;
static void ees_write(uint8_t data)
{
		if (m_addr >= sizeof(m_memory))
				m_addr = 0;
		
		if (m_addr != 0x2e)
		{
			m_memory[m_addr++] = data;
		}
		
		if (m_addr == (0x30 + m_memory[0x2f]))
		{
				if(m_memory[0x2f])
				{
					  memset(m_dfu.recv_buffer, 0 , 129);
						memcpy(m_dfu.recv_buffer, &m_memory[0x30] , m_memory[0x2f]);
					//for(int i = 0;i<m_memory[0x2f];i++)
					  //NRF_LOG_INFO(" m_dfu.recv_buffer:0x%02x",m_dfu.recv_buffer[i]);
					
						m_dfu.recv_length= m_memory[0x2f];
                    
            //NRF_LOG_INFO(" m_memory:0x%02x,0x%02x,0x%02x,0x%02x",m_memory[0x2f],m_memory[0x30],m_memory[0x31],m_memory[0x32]);
						//NRF_LOG_INFO(" m_dfu.recv_buffer:0x%02x,0x%02x,0x%02x",m_dfu.recv_buffer[0],m_dfu.recv_buffer[1],m_dfu.recv_buffer[2]);
						nrf_dfu_serial_on_packet_received(&m_serial,(uint8_t const *)m_dfu.recv_buffer,m_dfu.recv_length);
						//NRF_LOG_INFO(" m_dfu.recv_length:%d, num = %d",m_memory[0x2f],++num);
						m_memory[0x2f] = 0x00;
                    
						uint8_t * p_rx_buf = nrf_balloc_alloc(&m_payload_pool);
						if (p_rx_buf == NULL)
						{
								NRF_LOG_ERROR("Failed to allocate buffer");
								return;
						}
						//NRF_LOG_INFO("Allocated buffer %x", p_rx_buf);
						// reset the slip decoding
						m_slip.p_buffer      = &p_rx_buf[OPCODE_OFFSET];
						m_slip.current_index = 0;
						m_slip.state         = SLIP_STATE_DECODING;
				}
		}
}

/**
 * @brief Start after the WRITE command.
 *
 * Function sets pointers for TWIS to receive data.
 * WRITE command does not write directly to memory array.
 * Temporary receive buffer is used.
 * @sa m_rxbuff
 */
static void ees_writeBegin(void)
{
		(void)nrf_drv_twis_rx_prepare(&m_twis, m_rxbuff, sizeof(m_rxbuff));
}

/**
 * @brief Finalize the WRITE command.
 *
 * Call this function when the write command is finished.
 * It sets a memory pointer and writes all received data to the memory array.
 */
static void ees_writeEnd(size_t cnt)
{
	
		if (cnt > (EEPROM_SIM_ADDRESS_LEN_BYTES - 1))
		{
				size_t n;
#if (EEPROM_SIM_ADDRESS_LEN_BYTES == 1)
				ees_setAddr(m_rxbuff[0]);
#endif

#if (EEPROM_SIM_ADDRESS_LEN_BYTES == 2)
				uint16_t rxbuff;

				if (TWI_ADDRESS_CONFIG == LITTLE_ENDIAN)
				{
						rxbuff = ((m_rxbuff[1] << 8 ) | (m_rxbuff[0]));
				}
				else
				{
						rxbuff = ((m_rxbuff[0] << 8 ) | (m_rxbuff[1]));
				}

				ees_setAddr(rxbuff);
#endif

//				NRF_LOG_INFO("cnt = %d",cnt);
				//NRF_LOG_INFO(" m_rxbuff[0]:0x%02x,m_rxbuff[1]:0x%02x,m_rxbuff[2]:0x%02x,m_rxbuff[3]:0x%02x,m_rxbuff[4]:0x%02x",m_rxbuff[0],m_rxbuff[1],m_rxbuff[2],m_rxbuff[3],m_rxbuff[4]);
				for (n=EEPROM_SIM_ADDRESS_LEN_BYTES; n<cnt; ++n)
				{
						ees_write(m_rxbuff[n]);
						//NRF_LOG_INFO(" m_rxbuff[%d]:0x%02x",n,m_rxbuff[n]);
				}
		}
}

/**
 * @brief Start after the READ command.
 *
 * Function sets pointers for TWIS to transmit data from the current address to the end of memory.
 */
static void ees_readBegin(void)
{
		uint32_t read_len = EEPROM_SIM_SEQ_READ_MAX_BYTES;
		volatile uint16_t temp_addr = m_addr;

		if (m_addr + EEPROM_SIM_SEQ_READ_MAX_BYTES > sizeof(m_memory))
		{
				read_len = EEPROM_SIM_SIZE - m_addr;
				temp_addr = 0;
		}

		(void) nrf_drv_twis_tx_prepare(&m_twis, m_memory + m_addr, read_len);
		m_addr = temp_addr;
		
		
		//NRF_LOG_INFO("twis_tx ees_readBegin 0x%02x", m_memory[0x2e]);
}

/**
 * @brief Finalize the READ command.
 *
 * Call this function when the read command is finished.
 * It adjusts the current m_addr pointer.
 * @param cnt Number of bytes read.
 */
static void ees_readEnd(size_t cnt)
{
		m_addr += cnt;
}


static void payload_free(void * p_buf)
{
    uint8_t * p_buf_root = (uint8_t *)p_buf - DATA_OFFSET; //pointer is shifted to point to data
    
    nrf_balloc_free(&m_payload_pool, p_buf_root);
}

static ret_code_t rsp_send(uint8_t const * p_data, uint32_t length)
{
    uint32_t slip_len;
    (void) slip_encode(m_rsp_buf, (uint8_t *)p_data, length, &slip_len);

    return nrf_drv_twis_tx_prepare(&m_twis, m_rsp_buf, slip_len);
}

/**
 * @brief Event processing.
 *
 *
 */
static void twis_event_handler(nrf_drv_twis_evt_t const * const p_event)
{
		switch (p_event->type)
		{
		case TWIS_EVT_READ_REQ:
				if (p_event->data.buf_req)
				{
						ees_readBegin();
				}
				break;
		case TWIS_EVT_READ_DONE:
			{
				ees_readEnd(p_event->data.tx_amount);
				
			}
				break;
		case TWIS_EVT_WRITE_REQ:
				if (p_event->data.buf_req)
				{
						ees_writeBegin();
				}
				break;
		case TWIS_EVT_WRITE_DONE:
				ees_writeEnd(p_event->data.rx_amount);
		
				//NRF_LOG_INFO(" Transfer completed. Received: %d bytes, %02x",p_event->data.rx_amount,m_rxbuff[0]);
				break;

		case TWIS_EVT_READ_ERROR:
			NRF_LOG_INFO(" TWIS_EVT_READ_ERROR");
		break;
		case TWIS_EVT_WRITE_ERROR:
			NRF_LOG_INFO(" TWIS_EVT_WRITE_ERROR");
		break;
		case TWIS_EVT_GENERAL_ERROR:
			NRF_LOG_INFO(" TWIS_EVT_GENERAL_ERROR");
				m_error_flag = true;
				break;
		default:
				break;
		}
}

/** @} */
uint32_t i2c_dfu_transport_init()
{
    uint32_t ret;
		uint32_t err_code;

    NRF_LOG_DEBUG("i2c_dfu_transport_init()");

    err_code = nrf_balloc_init(&m_payload_pool);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }

		if (m_active)
    {
				NRF_LOG_DEBUG("m_active return");
        return err_code;
    }
		
    uint8_t * p_rx_buf = nrf_balloc_alloc(&m_payload_pool);

    m_slip.p_buffer      =  &p_rx_buf[OPCODE_OFFSET];
    m_slip.current_index = 0;
    m_slip.buffer_len    = UART_SLIP_MTU;
    m_slip.state         = SLIP_STATE_DECODING;

    m_serial.rsp_func           = rsp_send;
    m_serial.payload_free_func  = payload_free;
    m_serial.mtu                = UART_SLIP_MTU;
    m_serial.p_rsp_buf          = &m_rsp_buf[NRF_UART_MAX_RESPONSE_SIZE_SLIP -
                                            NRF_SERIAL_MAX_RESPONSE_SIZE];

    m_config.slave_addr                = EEPROM_SIM_ADDR;
    m_config.scl_s                     = EEPROM_SIM_SCL_S;
    m_config.sda_s                     = EEPROM_SIM_SDA_S;
    m_config.flash_start_addr          = EEPROM_SIM_FLASH_ADDRESS;
	
    /* Initialize RAM with contents of flash */
    for (uint32_t n=0; n<EEPROM_SIM_SIZE; ++n)
    {
        m_memory[n] = *(uint8_t *)(m_config.flash_start_addr + n);
    }
    m_memory[0x2e] = 0xcc;
    m_addr = 0;


    const nrf_drv_twis_config_t config =
    {
        .addr               = {m_config.slave_addr, 0},
        .scl                = m_config.scl_s,
        .scl_pull           = NRF_GPIO_PIN_PULLUP,
        .sda                = m_config.sda_s,
        .sda_pull           = NRF_GPIO_PIN_PULLUP,
        .interrupt_priority = APP_IRQ_PRIORITY_HIGH,
				.p_context = &m_serial
    };
		
    /* Init TWIS */
    do
    {
        ret = nrf_drv_twis_init(&m_twis, &config, twis_event_handler);
        if (NRF_SUCCESS != ret)
        {
						NRF_LOG_INFO(" nrf_drv_twis_init error!");
            break;
        }

        nrf_drv_twis_enable(&m_twis);
    }while (0);
		
		m_active = true;
		
		//Start RX
		err_code = nrf_drv_twis_rx_prepare(&m_twis, m_rxbuff, m_length);
		APP_ERROR_CHECK(err_code);

    return ret;
}

static uint32_t i2c_dfu_transport_close(nrf_dfu_transport_t const * p_exception)
{
	if ((m_active == true) && (p_exception != &dfu_trans))
	{
		 //nrf_drv_twis_uninit(&m_twis);
		 m_active = false;
		
	}
    
    return NRF_SUCCESS;
}

bool eeprom_simulator_error_check(void)
{
    return m_error_flag;
}

uint32_t eeprom_simulator_error_get_and_clear(void)
{
    uint32_t ret = nrf_drv_twis_error_get_and_clear(&m_twis);
    m_error_flag = false;
    return ret;
}

#endif

